@@ -20,7 +20,7 @@ from numpy.core import numeric
 
 # Local modules:
 import uncertainties
-from uncertainties import umath
+from uncertainties import umath, deprecation
 
 from uncertainties import __author__
 
@@ -238,26 +238,31 @@ def wrap_array_func(func):
 ###############################################################################
 # Arrays
 
-# Vectorized creation of an array of variables:
-
-# ! Looking up uncertainties.Variable beforehand through '_Variable =
-# uncertainties.Variable' does not result in a significant speed up:
-
-_uarray = numpy.vectorize(lambda v, s: uncertainties.Variable(v, s),
-                          otypes=[object])
-
-def uarray((values, std_devs)):
+def uarray(nominal_values, std_devs=None):
     """
     Returns a NumPy array of numbers with uncertainties
     initialized with the given nominal values and standard
     deviations.
 
-    values, std_devs -- valid arguments for numpy.array, with
+    nominal_values, std_devs -- valid arguments for numpy.array, with
     identical shapes (list of numbers, list of lists, numpy.ndarray,
     etc.).
+
+    std_devs=None is only used for supporting legacy code, where
+    nominal_values can be the tuple of nominal values and standard
+    deviations.
     """
 
-    return _uarray(values, std_devs)
+    if std_devs is None:  # Obsolete, single tuple argument call
+        deprecation('uarray() should now be called with two arguments.')
+        (nominal_values, std_devs) = nominal_values
+        
+    return (numpy.vectorize(
+        # ! Looking up uncertainties.Variable beforehand through
+        # '_Variable = uncertainties.Variable' does not result in a
+        # significant speed up:
+        lambda v, s: uncertainties.Variable(v, s), otypes=[object])
+        (nominal_values, std_devs))
 
 ###############################################################################
 
@@ -556,18 +561,22 @@ class matrix(numpy.matrix):
     
     std_devs = std_devs
     
-def umatrix(*args):
+def umatrix(nominal_values, std_devs=None):
     """
     Constructs a matrix that contains numbers with uncertainties.
 
-    The input data is the same as for uarray(...): a tuple with the
-    nominal values, and the standard deviations.
+    The arguments are the same as for uarray(...): nominal values, and
+    standard deviations.
 
     The returned matrix can be inverted, thanks to the fact that it is
     a unumpy.matrix object instead of a numpy.matrix one.
     """
 
-    return uarray(*args).view(matrix)
+    if std_devs is None:  # Obsolete, single tuple argument call
+        deprecation('umatrix() should now be called with two arguments.')
+        (nominal_values, std_devs) = nominal_values
+            
+    return uarray(nominal_values, std_devs).view(matrix)
 
 ###############################################################################
 
